from sympy.printing.python import PythonPrinter
import keyword as kw
class ModelPrinter(PythonPrinter):
    def __init__(self, settings=None):
        super(ModelPrinter, self).__init__(settings)
        self.model_symbols = []
    
    def _print_ModelMatrixSymbol(self,expr):
        symbol = self._str(expr)
        index = int(expr.name.split('_')[-1])
        matrix = '_'.join(expr.name.split('_')[0:-1])
        if symbol not in self.model_symbols:
            self.model_symbols.append(symbol)
        return f'p.{matrix}[{index}]'
    
    def _print_ModelSymbol(self,expr):
        symbol = self._str(expr)
        if symbol not in self.model_symbols:
            self.model_symbols.append(symbol)
        return f'p.{expr.name}'

def model(expr,p,**settings):
    """Return Python interpretation of passed expression
    (can be passed to the exec() function without any modifications)"""
    printer = ModelPrinter(settings)
    exprp = printer.doprint(expr)

    result = ''
    # Returning found symbols and functions
    renamings = {}
    for symbolname in printer.symbols:
        newsymbolname = symbolname
        # Escape symbol names that are reserved python keywords
        if kw.iskeyword(newsymbolname):
            while True:
                newsymbolname += "_"
                if (newsymbolname not in printer.symbols and
                        newsymbolname not in printer.functions):
                    renamings[sympy.Symbol(
                        symbolname)] = sympy.Symbol(newsymbolname)
                    break
        result += newsymbolname + ' = Symbol(\'' + symbolname + '\')\n'

    for functionname in printer.functions:
        if functionname in [i.name for i in p.q]:
            continue
        newfunctionname = functionname
        # Escape function names that are reserved python keywords
        if kw.iskeyword(newfunctionname):
            while True:
                newfunctionname += "_"
                if (newfunctionname not in printer.symbols and
                        newfunctionname not in printer.functions):
                    renamings[sympy.Function(
                        functionname)] = sympy.Function(newfunctionname)
                    break
        result += newfunctionname + ' = Function(\'' + functionname + '\')\n'

    if renamings:
        exprp = expr.subs(renamings)
    exprp_str = printer._str(exprp)
    for i,q_i in enumerate(p.q):
        exprp_str = exprp_str.replace(f'Derivative({q_i.name}(t), t)',f'p.qd[{i}]').replace(f'{q_i.name}(t)',f'p.q[{i}]')
    result += 'e = ' + exprp_str
    return result


from sympy.utilities.lambdify import _EvaluatorPrinter
from sympy.core.compatibility import (exec_, is_sequence, iterable,
    NotIterable, builtins)
    
def doprint(self, funcname, args, expr):
        """Returns the function definition code as a string."""
        from sympy import Dummy,cse,Symbol

        funcbody = []

        if not iterable(args):
            args = [args]

        argstrs, expr = self._preprocess(args, expr)

        ## --------------- Addition -----------------
        replacments, exprs = cse(expr,symbols=(Symbol(f'rep_{i}')for i in range(10000)))
        if isinstance(expr,tuple):
            expr = tuple(exprs)
        elif isinstance(expr,list):
            expr = exprs
        else:
            expr = exprs[0]
        ## --------------- Addition -----------------

        # Generate argument unpacking and final argument list
        funcargs = []
        unpackings = []

        for argstr in argstrs:
            if iterable(argstr):
                funcargs.append(self._argrepr(Dummy()))
                unpackings.extend(self._print_unpacking(argstr, funcargs[-1]))
            else:
                funcargs.append(argstr)
        arg_vars = ', '.join(funcargs)
        funcsig = f'def {funcname}({arg_vars}):'

        # Wrap input arguments before unpacking
        funcbody.extend(self._print_funcargwrapping(funcargs))

        funcbody.extend(unpackings)

        ## --------------- Addition -----------------
        for variable, expression in replacments:
            funcbody.append(f'{variable} = {self._exprrepr(expression)}')
        ## --------------- Addition -----------------

        funcbody.append('return ({})'.format(self._exprrepr(expr)))

        funclines = [funcsig]
        funclines.extend('    ' + line for line in funcbody)

        return '\n'.join(funclines) + '\n'


import sympy as sym

def MassMatrix(m,I_xx=0,I_yy=0,I_zz=0,I_xy=0,I_xz=0,I_yz=0):
    M = sym.diag(m,m,m,I_xx,I_yy,I_zz)
    M[4,3]=M[3,4]=I_xy
    M[5,3]=M[3,5]=I_xz
    M[5,4]=M[4,5]=I_yz
    return M


import sympy as sym
import sympy.physics.mechanics as me
import os
from .model_parameters import ModelParameters,ModelSymbol,ModelMatrix

class DynamicModelParameters(ModelParameters):
    def __init__(self,DoFs):
        self.qs = DoFs
        self.q = sym.Matrix(me.dynamicsymbols(f'q:{DoFs}'))
        self.qd = sym.Matrix(me.dynamicsymbols(f'q:{DoFs}',1))
        self.qdd = sym.Matrix(me.dynamicsymbols(f'q:{DoFs}',2))

        # create state matrix
        self.x = sym.BlockMatrix([[self.q],[self.qd]]).as_explicit()
        super().__init__()

    def print_python(self):
        ignore = ['qs','q','qd','qdd']
        string = f'p = ma.DynamicModelParameters({self.qs})'
        for name,value in vars(self).items():
            if name not in ignore:
                if isinstance(value, ModelSymbol):
                    if callable(value.value):
                        string += f'\np.{name} = ma.ModelSymbol(string=\'{value.name}\')'
                    else:
                        string += f'\np.{name} = ma.ModelSymbol(value={value.value}, string=\'{value.name}\')'
                elif isinstance(value, ModelMatrix):
                    if callable(value.value):
                        string += f'\np.{name} = ma.ModelMatrix(string=\'{value._matrix_symbol}\', length={len(value)})'
                    else:
                        string += f'\np.{name} = ma.ModelMatrix(value={value.value}, string=\'{value._matrix_symbol}\', length={len(value)})'
                elif isinstance(value,sym.Symbol):
                    string += f'\np.{name} = Symbol(\'{value.name}\')'
        return string

    def to_file(self,filename,file_dir=''):
        string = 'import moyra as ma\nfrom sympy import *\n\ndef get_p():\n\t' + self.print_python().replace('\n','\n\t') + '\n\treturn p\n'
        with open(os.path.join(file_dir,filename),'w') as file:
            file.write(string)


from .model_parameters import ModelParameters, ModelMatrix, ModelSymbol,ModelMatrixSymbol, ModelExpr, ModelValue
from .dynamic_model_parameters import DynamicModelParameters
from .helper_funcs import linearise_matrix, extract_eigen_value_data, Vee, Vee4, Wedge, Wedge4, partial_wrt_t, trigsimp
from .symbolic_model import SymbolicModel
from .numeric_model import NumericModel

from . import frames as frames
from . import elements as elements
from . import forces as forces

# monkey patch lambdify to use common sub expression reduction
from sympy.utilities.lambdify import _EvaluatorPrinter
from .lambdify_extension import doprint
_EvaluatorPrinter.doprint = doprint


from .external_force import ExternalForce
from .aero_force import AeroForce
from .gravity_force import GravityForce
from .composite_force import CompositeForce
from .custom_force import CustomForce
from .continuous_grav_force import ContinuousGravityForce



from . import ExternalForce

class CustomForce(ExternalForce):

    def __init__(self,Forcingfunction):
        self.q_func = Forcingfunction

    def Q(self):
        return None
    
    def subs(self,*args):
        return self

    def msubs(self,*args):
        return self

    def cancel(self):
        return self

    def expand(self):
        return self

    def integrate(self,*args):
        return self

    def linearise(self,x,x_f):
        return self

    def lambdify(self,params):
        return self.q_func   

        



from functools import cache
import warnings
import sympy as sym
from sympy.abc import t

from moyra.symbolic_model import SymbolicModel

class BaseElement:
    def __init__(self,q,name="default") -> None:
        self._q = q
        self._elementname = name

    @property
    def q(self):
        return self._q

    @property
    @cache
    def qd(self):
        return self.q.diff(t)
    
    @property
    @cache
    def qdd(self):
        return self.q.diff(t,2)

    @property
    def ke(self):
        warnings.warn(f'ke not implemented in the ele {self}')
        return 0

    @property
    def pe(self):
        warnings.warn(f'pe not implemented in the ele {self}')
        return 0

    @property
    def rdf(self):
        warnings.warn(f'rdf not implemented in the ele {self}')
        return 0

    @property
    def M(self):
        warnings.warn(f'M not implemented in the ele {self}')
        return sym.zeros(len(self.q))

    def to_symbolic_model(self, legacy=False):
        Lag = sym.Matrix([self.ke-self.pe])
        D = sym.Matrix([self.rdf])
        # legacy method is a lot slower but can produce more compact results
        Q_v = (self.M.diff(t))*self.qd if not legacy else Lag.jacobian(self.qd).diff(t).T
        term_2 = Lag.jacobian(self.q).T
        term_3 = D.jacobian(self.qd).T
        f = Q_v - term_2 + term_3
        return SymbolicModel(self.q,self.M,f,self.ke,self.pe)      

    @property
    def elementname(self):
        return self._elementname
    @elementname.setter
    def elementname(self,value):
        self._elementname = value

    def __str__(self):
        return f'{self.elementname}:{self.__class__.__name__}'


import sympy as sym
import sympy.physics.mechanics as me

class ModelValue:
    """
    Base class to inject a value onto sympy classes
    """
    def __init__(self,value=0,**kwarg):
        super().__init__(**kwarg)
        self.value = value
        self._dependent = False      
    
    def __call__(self,x,t):
        return self._GetValue(x,t)

    def _GetValue(self,x,t):
        if callable(self.value):
            return self.value(x,t)
        else:
            return self.value

    def GetSub(self,t,x):
        return self(t,x)
        
class ModelSymbol(sym.Symbol,ModelValue):
    """
    Wrapper for Sympy Symbol, to inject it with a value attribute
    """
    def __init__(self,string,**kwarg):
        super().__init__(**kwarg)
    def __new__(cls,string,**kwarg):
        return super().__new__(cls,string)
    def __eq__(self,other):
        if isinstance(other,sym.Symbol):
            return other.name == self.name
    def __hash__(self):
        return hash(sym.Symbol(self.name))
    def _octave(self,printer):
        return f'{self.name}'

class ModelMatrixSymbol(ModelSymbol):
    def __init__(self,string,**kwarg):
        self._index = int(string.split('_')[-1])
        self._matrix = '_'.join(string.split('_')[0:-1])
        super().__init__(string,**kwarg)
    def __new__(cls,string,**kwarg):
        return super().__new__(cls,string,**kwarg)
    def _octave(self,printer):
        return f'{self._matrix}({self._index+1})'

class ModelMatrix(sym.Matrix,ModelValue):
    """
    Wrapper for Sympy Matrix, to inject it with a value attribute
    """
    def __init__(self,string,length,**kwarg):
        if "value" not in kwarg:
            kwarg["value"] = [0]*length
        super().__init__(**kwarg)
        self._matrix_symbol = string
    def __new__(cls,string,length,**kwargs):
        return super().__new__(cls,sym.symbols(f'{string}_:{length}',cls=ModelMatrixSymbol))
    def __setattr__(self,name,value):
        if name == "value":
            if value is not None:
                r, c = self.shape
                if len(value) != r*c:
                    raise ValueError(f'Model Matrix value length, {len(value)}, must be the same length as the symbolic matrix, {self.shape}.')
        object.__setattr__(self, name, value)



class ModelExpr(sym.Symbol,ModelValue):
    def __init__(self,string,func,**kwarg):
        self.expr_func = func
        super().__init__(**kwarg)

    def _GetValue(self,t,x):
        return self.expr_func(t,x)

    def __new__(cls,string,**kwarg):
        return super().__new__(cls,string)

    def GetSub(self,t,x):
        return self.value
class ModelParameters:    
    
    def GetTuple(self,ignore=[]):
        return tuple(var for name,var in vars(self).items() if isinstance(var,ModelValue) and name not in ignore and var not in ignore)
    
    def GetSubs(self,t,x,ignore=[]):
        sub_dependent_dict = {}
        sub_dict = {}
        # put dependent substitions in first
        for name,var in vars(self).items():
            if isinstance(var,ModelValue) and name not in ignore and var not in ignore:
                if isinstance(var,ModelMatrix):
                    for i in range(len(var)):
                        sub_dict[var[i]] = var.GetSub(t,x)[i]
                else:
                    if var._dependent:
                        sub_dependent_dict[sym.Symbol(var.name)] = var.GetSub(t,x)
                    else:
                        sub_dict[sym.Symbol(var.name)] = var.GetSub(t,x)
        # sub in values for dependent subsitutions
        for key,value in sub_dependent_dict.items():
            sub_dependent_dict[key] = value.subs(sub_dict)
        #combine dictionaries
        tot_sub_dict = {**sub_dict,**sub_dependent_dict}
        # return a dictionary with all keys changed to symbols
        return tot_sub_dict#{sym.Symbol(k.name):v for k,v in tot_sub_dict.items()}
    
    def GetNumericTuple(self,x,t,ignore=[]):
        return tuple(var(x,t) for name,var in vars(self).items() if isinstance(var,ModelValue) and name not in ignore and var not in ignore)

    def to_matlab_class(self,class_name = "Parameters",file_dir='', ignore=[], base_class = None):
        import os.path
        # create a dict of all required params and values
        params = {}
        for name,var in vars(self).items():
            if name not in ignore and var not in ignore:
                if isinstance(var,ModelSymbol):
                    params[var.name] = var.value
                if isinstance(var,ModelMatrix):
                    params[name] = var.value
        # convert to matlab class string
        cn = class_name if base_class is None else class_name+f" < {base_class}"
        classdef = f'classdef {cn}'
        params = '\n\t\t'.join([ f'{key} = {value}' for key,value in params.items()]).replace('{','').replace('}','')
        class_string = classdef + '\n\tproperties\n\t\t' + params + '\n\tend\nend'
        # save to file
        with open(os.path.join(file_dir,class_name + '.m'),'w') as file:
            file.write(class_string)


import sympy as sym
from sympy.abc import t
import sympy.physics.mechanics as me

class BaseReferenceFrame:

    def __init__(self,A=None,R=None):
        self._A = sym.eye(3) if A is None else sym.Matrix(A)
        self._R = sym.zeros(3,1) if R is None else sym.Matrix(R)

    A = property(lambda self: self._A)
    R = property(lambda self: self._R)    

    def transform_point(self,p):
        return self.A*sym.Matrix(list(p))+self.R

    def transform_global_point(self,p):
        return self.A.T*(sym.Matrix(list(p))-self.R)

    def transform_vector(self,v):
        return self.A*sym.Matrix(list(v))

    def transform_global_vector(self,v):
        return self.A.T*sym.Matrix(list(v))


import sympy as sym
import numpy as np
from . import ExternalForce

class CompositeForce(ExternalForce):
    """A Class used to represent the combination of multiple ExternalForce objects"""

    def __init__(self,forces):
        """constructor for composite force, where 'forces' is a enumerable of ExternalForce objects"""
        self.forces = forces

    def Q(self):
        filt = filter(None,(f.Q() for f in self.forces))
        return sum(filt,next(filt))

    def subs(self,*args):
        return CompositeForce([force.subs(*args) for force in self.forces])

    def msubs(self,*args):
        return CompositeForce([force.msubs(*args) for force in self.forces])

    def cancel(self):
        return CompositeForce([force.cancel() for force in self.forces])

    def expand(self):
        return CompositeForce([force.expand() for force in self.forces])

    def linearise(self,x,x_f):
        return CompositeForce([force.linearise(x,x_f) for force in self.forces])

    def lambdify(self,params):
        force_funcs = list(filter(None,[force.lambdify(params) for force in self.forces]))
        return CallableCompositeForce(force_funcs)

class CallableCompositeForce:
    def __init__(self,force_funcs):
        self.force_funcs = force_funcs
    
    def __call__(self,*params):
        return sum((force_func(*params) for force_func in self.force_funcs))


import sympy as sym
from . import ExternalForce
from ..helper_funcs import linearise_matrix
import sympy.physics.mechanics as me

class AeroForce(ExternalForce):
    """A Class used to represent aerodynamic forces"""
    @classmethod
    def PerUnitSpan(cls,p,Transform,C_L,alphadot,M_thetadot,e,alpha_zero=0, root_alpha=0, delta_alpha=0,stall_angle=0.24,c_d_max = 1,w_g = 0,V=None,c=None,linear=False,z_inverted=False,c_l_func=None,C_M=0,c_m_func=None,unsteady=False):
        """ A static method to create an aerodynamic force per unit span

        ...
        Parameters
        ----------
        p - an instance of a 'ModelParameters' class
        Transform - an instance of a 'HomogenousTransform' class describing the coordinate system the force is applied in
        C_L - the 2D lift curve slope
        alphadot - unsteady aerodynamic coeffiecent (see Cooper & Wright)
        M_thetadot - unsteady aerodynamic coeffiecent (see Cooper & Wright)
        e - chord wing eccentrity
        rootAlpha - root angle of attack of the element
        alpha_zero - angle attack for zero lift (camber of aerofoil)
        
        Optional Parameters
        -------------------

        stall_angle - angle in radians at which the wing stalls, defining the lift curve slope. If zero assume linear (default = 0)
        c_d_max - max coeefgficent of drag (default = 1)
        w_g - gust velocity (default = 0)
        V - onset wind velocity (default p.V)
        c - chord (default p.c)
        linear - if true normal force is a combination of lift and drag (default false)
        z_inverted - if true lift acts in the oppiste direction to the z axis
        """
        ## force per unit length will following theredosons pseado-steady theory
        (Q,dAlpha) = cls.get_Q(p,Transform,C_L,alphadot,M_thetadot,e,
                            alpha_zero,root_alpha,delta_alpha,stall_angle,c_d_max,w_g,
                            V,c,linear,z_inverted,c_l_func,C_M,c_m_func,unsteady)
        return cls(Q,dAlpha)
        
    def __init__(self,Q,dAlpha):
        self.dAlpha = dAlpha
        super().__init__(Q) 

    @staticmethod
    def _trigsimp(expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr))))      

    def linearise(self,x,x_f):
        Q_lin = linearise_matrix(self.Q(),x,x_f)
        dAlpha_lin = linearise_matrix(self.dAlpha,x,x_f)
        return AeroForce(Q_lin,dAlpha_lin)
    
    def subs(self,*args):
        return AeroForce(self._Q.subs(*args),self.dAlpha.subs(*args))

    def msubs(self,*args):
        return AeroForce(me.msubs(self._Q,*args),me.msubs(self.dAlpha,*args))

    def integrate(self,*args):
        return AeroForce(self._Q.integrate(*args),self.dAlpha)

    @staticmethod
    def get_dAlpha(p,BodyJacobian,alphadot,alpha_zero=0, root_alpha=0, delta_alpha=0, w_g = 0,V=None,c=None,z_inverted=False,unsteady=False):
        if c is None:
            c=p.c

        v_z_eff = (BodyJacobian*p.qd)[2]
        if z_inverted:
            v_z_eff *= -1
        if V is None:
            V = -(BodyJacobian*p.qd)[0]
        
        alpha = delta_alpha + sym.atan(-v_z_eff/V) + w_g/V
        if unsteady:
            c1 = 0.5*p.AR/(2.32+p.AR)
            c2 = 0.181+0.772/p.AR
            G = -c1*c2*p.k/(p.k**2+c2**2)
            F = 1-c1*p.k**2/(p.k**2+c2**2)
            lag_alpha =  p.AR/(2+p.AR)*(F*(alpha)+p.c/(2*V*p.k)*G*alphadot)
        else:
            lag_alpha =  alpha
        return alpha_zero + root_alpha + lag_alpha

    def get_apparent_mass(p,BodyJacobian,C_L,V=None,c=None,linear=False,c_l_func=None):
        """
        see the class method PerUnitSpan for a explaination of terms
        """
        if c is None:
            c=p.c
        if V is None:
            V = -(BodyJacobian*p.qd)[0]
        v = (BodyJacobian*p.qd)[2]
        return p.rho*C_L*p.c**2*v.diff(me.dynamicsymbols._t)/8

    def get_F_n(p,BodyJacobian,dAlpha,C_L,alpha_zero=0, stall_angle=0.24,c_d_max = 1,V=None,c=None,linear=False,c_l_func=None):
        """
        see the class method PerUnitSpan for a explaination of terms
        """
        if c is None:
            c=p.c
        if V is None:
            V = -(BodyJacobian*p.qd)[0]

        v_z_eff = (BodyJacobian*p.qd)[2]
        V_rel = sym.sqrt(v_z_eff**2 + V**2)
        # Calculate the lift force
        dynamicPressure = sym.Rational(1,2)*p.rho*V_rel**2

        # Calculate C_L curve
        if c_l_func is not None:
            c_l = c_l_func(dAlpha)
        else:
            if stall_angle == 0:
                c_l = C_L*dAlpha
            else:
                c_l = C_L*(1/p.clip_factor*sym.ln((1+sym.exp(p.clip_factor*(dAlpha+stall_angle)))/(1+sym.exp(p.clip_factor*(dAlpha-stall_angle))))-stall_angle)

        if linear:
            c_n = c_l
        else:
            c_d = c_d_max*sym.Rational(1,2)*(1-sym.cos(2*dAlpha))
            ang = dAlpha - alpha_zero
            c_n = c_l*sym.cos(ang)+c_d*sym.sin(ang)

        ## joint torques for lift are calculated in a frame aligned with the chordwise velocity direction
        return  dynamicPressure*c*c_n
        
    def get_moment(p,BodyJacobian,dAlpha,F_n,C_L,alphadot,M_thetadot,e,V=None,c=None,C_M=0,c_m_func=None,unsteady=False):
        if c_m_func is not None:
            c_m = c_m_func(dAlpha)
        else:
            c_m = C_M*dAlpha
        if V is None:
            V = -(BodyJacobian*p.qd)[0]   
        dynamicPressure = sym.Rational(1,2)*p.rho*V**2
        M_w = -F_n*e*c + dynamicPressure*c*-c_m# Moment due to lift
        M_w += dynamicPressure*c**2*(M_thetadot*alphadot*c/(sym.Integer(4)*V))
        return M_w

    def get_Q(p,Transform,C_L,alphadot,M_thetadot,e,alpha_zero=0, root_alpha=0, delta_alpha=0, stall_angle=0.24,c_d_max = 1,w_g = 0,V=None,c=None,linear=False,z_inverted=False,c_l_func=None,C_M=0,c_m_func=None,unsteady=False):
        """
        see the class method PerUnitSpan for a explaination of terms
        """
        BodyJacobian = sym.simplify(AeroForce._trigsimp(Transform.BodyJacobian(p.q)))
        dAlpha = AeroForce.get_dAlpha(p,BodyJacobian,alphadot,alpha_zero, root_alpha, delta_alpha, w_g,V,c,z_inverted,unsteady)
        F_n = AeroForce.get_F_n(p,BodyJacobian,dAlpha,C_L,alpha_zero, stall_angle,c_d_max,V,c,linear,c_l_func)
        M_w = AeroForce.get_moment(p,BodyJacobian,dAlpha,F_n,C_L,alphadot,M_thetadot,e,V,c,C_M,c_m_func,unsteady)
        if c is None:
            c=p.c
        if V is None:
            V = -(BodyJacobian*p.qd)[0]
        # Calculate the lift force

        ## joint torques for lift are calculated in a frame aligned with the chordwise velocity direction
        wrench_lift = sym.Matrix([0,0,-F_n,0,0,0]) if z_inverted else sym.Matrix([0,0,F_n,0,0,0])
        Q_L = BodyJacobian.T*wrench_lift

        ## joint torques for moment are calculated in a frame aligned with the shord of the wing
        BodyJacobian = sym.simplify(AeroForce._trigsimp(Transform.R_y(delta_alpha).BodyJacobian(p.q)))
        wrench_moment = sym.Matrix([0,0,0,0,M_w,0])
        Q_M = BodyJacobian.T*wrench_moment

        # combine forces
        Q = Q_L + Q_M
        return (Q,dAlpha)





import sympy as sym
import numpy as np
from scipy.linalg import eig
from .helper_funcs import linearise_matrix

class NumericModel:

    def __init__(self,p,M,f,T,U,ExtForces = None):
        from sympy.abc import t
        """Initialise a Symbolic model of the form 
        $M\ddot{q}+f(\dot{q},q,t)-ExtForces(\dot{q},q,t) = 0$

        with the Symbolic Matricies M,f,and Extforces
        """
  
        #generate lambda functions
        tup = p.GetTuple()
        
        # External Forces
        self.ExtForces = ExtForces.lambdify((p.x,tup,t)) if ExtForces is not None else lambda x,tup,t : 0
        # Mass Matrix Eqn
        self.M = sym.lambdify((p.x,tup),M,"numpy")
        #func eqn
        self.f = sym.lambdify((p.x,tup),f,"numpy")
        # potential energy function
        self.pe = sym.lambdify((p.x,tup),U,"numpy")
        # kinetic energy function
        self.ke = sym.lambdify((p.x,tup),T,"numpy")

    @classmethod
    def from_SymbolicModel(cls,p,sm):
        return cls(p,sm.M,sm.f,sm.T,sm.U,sm.ExtForces)


    def deriv(self,x,tup,t):
        try:
            external = self.ExtForces(x,tup,t)
            accels = np.linalg.inv(self.M(x,tup))@(-self.f(x,tup)+external)
        except ZeroDivisionError:
            accels = np.linalg.inv(self.M(x,tup))@(-self.f(x,tup))        

        state_vc = np.append(x[int(len(x)/2):],accels)
        return tuple(state_vc)

    def energy(self,x,tup,t):
        return self.ke(x,tup) + self.pe(x,tup)


import sympy as sym
import sympy.physics.mechanics as me
from inspect import getsource
from ..helper_funcs import linearise_matrix


class ExternalForce:

    @classmethod
    def zero_force(cls,dofs):
        Q = sym.Matrix([0]*dofs)
        return cls(Q)

    @classmethod
    def body_force(cls,q,frame,Fx=0,Fy=0,Fz=0,Mx=0,My=0,Mz=0,simplify=True):
        BodyJacobian = frame.BodyJacobian(q)
        BodyJacobian = sym.simplify(cls._trigsimp(BodyJacobian)) if simplify else BodyJacobian
        wrench = sym.Matrix([Fx,Fy,Fz,Mx,My,Mz])
        return cls(BodyJacobian.T*wrench) 

    @staticmethod
    def _trigsimp(expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr)))) 

    def __init__(self,Q = None):
        self._Q = Q

    def __mul__(self,other):
        return ExternalForce(self._Q*other)

    def __add__(self,other):
        if isinstance(other, ExternalForce):
            return ExternalForce(self._Q+other._Q)
        else:
            return ExternalForce(self._Q+other)

    def Q(self):
        return self._Q

    def subs(self,*args):
        return ExternalForce(self._Q.subs(*args))

    def msubs(self,*args):
        return ExternalForce(me.msubs(self._Q,*args))

    def cancel(self):
        return ExternalForce(sym.cancel(self._Q))  

    def expand(self):
        return ExternalForce(sym.expand(self._Q))  

    def integrate(self,*args):
        return ExternalForce(self._Q.integrate(*args))

    def linearise(self,x,x_f):
        return ExternalForce(linearise_matrix(self.Q(),x,x_f))

    def lambdify(self,params):
        if self._Q is None:
            return None
        return sym.lambdify(params,self._Q ,"numpy")




    








import sympy as sym
import numpy as np
import sympy.physics.mechanics as me
from sympy.abc import t

def Vee4(E):
    val = sym.Matrix([[0]]*6)
    val[:3,0] = E[:3,3]
    val[3,0] = E[2,1]
    val[4,0] = E[0,2]
    val[5,0] = E[1,0]
    return val
    
def Vee(S):
    return sym.Matrix([S[2,1],S[0,2],S[1,0]])

def Wedge4(V):
    val = sym.Matrix([[0]*4]*4)
    val[:3,3] = V[:3,0]   
    val[0,1] = -V[-1]
    val[1,0] = V[-1]
    val[2,0] = -V[-2]
    val[0,2] = V[-2]
    val[1,2] = -V[-3]
    val[2,1] = V[-3]
    return val

def Wedge(V):
    val = sym.Matrix([[0]*3]*3)
    val[0,1] = -V[2]
    val[1,0] = V[2]
    val[2,0] = -V[1]
    val[0,2] = V[1]
    val[1,2] = -V[0]
    val[2,1] = V[0]
    return val

def partial_wrt_t(expr,q):
    qs = len(q)
    qd = q.diff(t)
    # insert  dummy variables
    U = sym.Matrix(sym.symbols(f'u_:{qs*2}'))
    l = dict(zip([*q,*qd],U))
    expr = me.msubs(expr,l)
    return expr.diff(t)

def trigsimp(expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr)))) 

def linearise_matrix(M,x,x_f):
    # reverse order of states to ensure velocities are subbed first
    x_subs = {x[i]:x_f[i] for i in range(len(x))}

    # get the value of M at the fixed point
    M_f = me.msubs(M,x_subs)

    # add a gradient term for each state about the fixed point
    
    for i,xi in enumerate(x):
        M_f += me.msubs(M.diff(xi),x_subs)*(xi-x_f[i])
    return M_f

def extract_eigen_value_data(evals,evecs,margin=1e-9,sortby=None):       
    # get unique eigen values
    unique = []
    unique_vecs = []
    for idx,val in enumerate(evals):
        if np.iscomplex(val):
            # check the complex conjugate is not already in the list
            if not any(np.isclose(np.conj(val),unique)):
                unique.append(val)
                unique_vecs.append(evecs[:,idx])
        else:
            # and real poles straight away
            unique.append(val)
            unique_vecs.append(evecs[:,idx].tolist())

    # Generate data
    real = np.real(unique)
    imag = np.imag(unique)
    F = np.where(np.iscomplex(unique),np.abs(unique)/(2*np.pi),0)
    D = np.where(np.iscomplex(unique),np.cos(np.angle(unique)),np.NaN)
    S = np.max(real)<=margin

    # got order to be sorted in
    if sortby == 'F':
        ind = np.argsort(F)
    elif sortby == 'D':
        ind = np.argsort(D)
    else:
        ind = range(len(unique)) 

    # place data in a dict and sort
    res = []
    Mode = 0
    for i in ind:
        res_dict = {}
        res_dict['Real'] = real[i]
        res_dict['Imag'] = imag[i]
        res_dict['Frequency'] = F[i]
        res_dict['Damping'] = D[i]
        res_dict['Stable'] = S
        res_dict['Eigen Vector'] = unique_vecs[i]
        res_dict['Mode'] = Mode
        Mode += 1
        res.append(res_dict)

    return res



import re
import os
import pickle
import sympy as sym
import numpy as np
from scipy.linalg import eig
import sympy.physics.mechanics as me
from sympy.physics.vector.printing import vpprint, vlatex
from sympy.abc import x,y,t

from moyra.forces.external_force import ExternalForce
from .helper_funcs import linearise_matrix, partial_wrt_t
from .numeric_model import NumericModel
import pickle
from .printing import model as print_model
from .model_parameters import ModelSymbol, ModelMatrix,ModelMatrixSymbol, ModelValue
from time import time, ctime
from collections.abc import Iterable
from sympy.abc import t

class SymbolicModel:
    """
    An instance of a folding wing tip model using assumed shapes.

    Required inputs are:
        generalisedCoords - array of the generalised coordinate symbols 
            (must be dynamic symbols)
        z_w,z_t,alpha_w,alpha_t - sympy expressions of the z and alpha postion
            of the wing and FWT
        FwtParameters - instance of the FwtParameters class (with the symbols 
            used in the above expressions)
        thetaIndex - index of theta (hinge angle) in generalisedCoords 
            (so energy equation knows which one) if no theta coordinate leave
            as 'None'
    """
    @classmethod
    def FromElementsAndForces(cls,q,Elements, ExtForces=None, C=None):
        """
        Create a symbolic Model instance from a set Elements and external forces
        """
        # Calc K.E, P.E and Rayleigh Dissaptive Function
        T = U = D = sym.Integer(0)
        qs = len(q)
        M = sym.zeros(qs)
        f = sym.zeros(qs,1)
        sm = cls(q,M,f,T,U)
        # add K.E for each Rigid Element
        Elements = Elements if isinstance(Elements,Iterable) else [Elements]
        for i,ele in enumerate(Elements):
            print(f'Generating EoM for Element {i+1} out of {len(Elements)} - {ele}')
            sm += ele.to_symbolic_model()
        
        return cls(q,sm.M,sm.f,sm.T,sm.U,ExtForces,C)

    @classmethod
    def FromElementsAndForces_2(cls,q,Elements, ExtForces=None, C=None):
        """
        Create a symbolic Model instance from a set Elements and external forces
        """
        # Calc K.E, P.E and Rayleigh Dissaptive Function
        T = U = D = sym.Integer(0)
        qs = len(q)
        M = sym.zeros(qs)
        f = sym.zeros(qs,1)
        sm = cls(q,M,f,T,U)
        # add K.E for each Rigid Element
        Elements = Elements if isinstance(Elements,Iterable) else [Elements]
        for i,ele in enumerate(Elements):
            print(f'Generating EoM for Element {i+1} out of {len(Elements)} - {ele}')
            sm += ele.to_symbolic_model_2()
        
        return cls(q,sm.M,sm.f,sm.T,sm.U,ExtForces,C)

    def __init__(self,q,M,f,T,U,ExtForces = None,C = sym.Matrix([])):
        """Initialise a Symbolic model of the form 
        $M\ddot{q}+f(\dot{q},q,t)-ExtForces(\dot{q},q,t) = 0, with constraints C=0$

        with the Symbolic Matricies M,f,and Extforces
        """
        self._M = M
        self._f = f
        self._T = T
        self._U = U
        self._q = q
        self._ExtForces = ExtForces if ExtForces is not None else ExternalForce.zero_force(f.shape[0])
        self._C = C

    M = property(lambda self:self._M)
    f = property(lambda self:self._f)
    T = property(lambda self:self._T)
    U = property(lambda self:self._U)
    q = property(lambda self:self._q)
    qd = property(lambda self:self.q.diff(t))
    qdd = property(lambda self:self.q.diff(t,2))
    qs = property(lambda self:len(self.q))
    ExtForces = property(lambda self:self._ExtForces)
    C = property(lambda self:self._C)

    def __add__(self,other):
        if not isinstance(other,SymbolicModel):
            raise TypeError('other must be of type SymbolicModel')
        idx = [i for i, e in enumerate(self.q) if e in other.q]
        
        # add mass matricies
        M = sym.Matrix(self.M)
        for i,m_i in enumerate(idx):
            for j,m_j in enumerate(idx):
                M[m_i,m_j] += other.M[i,j]

        # add forces matricies
        f = sym.Matrix(self.f)
        Q = self.ExtForces.Q()
        Q_other = other.ExtForces.Q()
        for i,m_i in enumerate(idx):
                f[m_i] += other.f[i]
                Q[m_i] += Q_other[i]

        T = self.T + other.T
        U = self.U + other.U
        C = sym.Matrix([*self.C,*other.C])

        return SymbolicModel(self.q,M,f,T,U,ExternalForce(Q),C)

    def cancel(self):
        """
        Creates a new instance of a Symbolic model with the cancel simplifcation applied
        """
        ExtForces = self.ExtForces.cancel() if self.ExtForces is not None else None

        # handle zero kinetic + pot energies
        T = self.T if isinstance(self.T,int) else sym.cancel(self.T)
        U = self.U if isinstance(self.U,int) else sym.cancel(self.U)
        C = self.C if self.C is None else sym.cancel(self.C)
        return SymbolicModel(self.q,sym.cancel(self.M),sym.cancel(self.f),
                            T,U,ExtForces,C)

    def expand(self):
        """
        Creates a new instance of a Symbolic model with the cancel simplifcation applied
        """
        ExtForces = self.ExtForces.expand() if self.ExtForces is not None else None

        # handle zero kinetic + pot energies
        T = self.T if isinstance(self.T,int) else sym.expand(self.T)
        U = self.U if isinstance(self.U,int) else sym.expand(self.U)
        C = self.C if self.C is None else sym.expand(self.C)
        return SymbolicModel(self.q,sym.expand(self.M),sym.expand(self.f),
                            T,U,ExtForces,C)


    def subs(self,*args):
        """
        Creates a new instance of a Symbolic model with the substutions supplied
         in args applied to all the Matricies
        """
        ExtForces = self.ExtForces.subs(*args) if self.ExtForces is not None else None

        # handle zero kinetic + pot energies
        T = self.T if isinstance(self.T,int) else self.T.subs(*args)
        U = self.U if isinstance(self.U,int) else self.U.subs(*args)
        C = self.C if self.C is None else self.C.subs(*args)
        return SymbolicModel(self.q,self.M.subs(*args),self.f.subs(*args),
                            T,U,ExtForces,C)

    def msubs(self,*args):
        """
        Creates a new instance of a Symbolic model with the substutions supplied
         in args applied to all the Matricies
        """
        ExtForces = self.ExtForces.msubs(*args) if self.ExtForces is not None else None

        # handle zero kinetic + pot energies
        T = self.T if isinstance(self.T,int) else me.msubs(self.T,*args)
        U = self.U if isinstance(self.U,int) else me.msubs(self.U,*args)
        C = self.C if self.C is None else me.msubs(self.C,*args)
        return SymbolicModel(self.q, me.msubs(self.M,*args),me.msubs(self.f,*args),
                            T,U,ExtForces,C)

    def linearise(self,x_f):
        """
        Creates a new instance of the symbolic model class in which the EoM have been 
        linearised about the fixed point p.q_0
        """
        # Calculate Matrices at the fixed point
        # (go in reverse order so velocitys are subbed in before positon)

        # get the full EoM's for free vibration and linearise

        x = [*self.q,*self.qd]
        x_subs = {x[i]:x_f[i] for i in range(len(x))}

        M_lin = me.msubs(self.M,x_subs)

        f_lin = linearise_matrix(self.f,x,x_f)
        T_lin = linearise_matrix(self.T,x,x_f)
        U_lin = linearise_matrix(self.U,x,x_f)

        # Linearise the External Forces
        extForce_lin = self.ExtForces.linearise(x,x_f) if self.ExtForces is not None else None

        # create the linearised model and return it
        return SymbolicModel(self.q,M_lin,f_lin,T_lin,U_lin,extForce_lin)

    def extract_matrices(self):
        """
        From the current symbolic model extacts the classic matrices A,B,C,D,E as per the equation below
        A \ddot{q} + B\dot{q} + Cq = D\dot{q} + Eq

        THE SYSTEM MUST BE LINEARISED FOR THIS TO WORK
        """
        A = self.M
        D = self.f.jacobian(self.qd)
        E = self.f.jacobian(self.q)
        B = self.ExtForces.Q().jacobian(self.qd)
        C = self.ExtForces.Q().jacobian(self.q)
        return A,B,C,D,E

    def free_body_eigen_problem(self):
        """
        gets the genralised eigan matrices for the free body problem.
        They are of the form:
            |   I   0   |       |    0    I   |
        M=  |   0   M   |   ,K= |   -C   -B   |
        such that scipy.linalg.eig(K,M) solves the problem 

        THE SYSTEM MUST BE LINEARISED FOR THIS TO WORK
        """
        M = sym.eye(self.qs*2)
        M[-self.qs:,-self.qs:]=self.M

        K = sym.zeros(self.qs*2)
        K[:self.qs,-self.qs:] = sym.eye(self.qs)
        K[-self.qs:,:self.qs] = -self.f.jacobian(self.q)
        K[-self.qs:,-self.qs:] = -self.f.jacobian(self.qd)
        return K,M

    def gen_eigen_problem(self):
        """
        gets the genralised eigan matrices for use in solving the frequencies / modes. 
        They are of the form:
            |   I   0   |       |    0    I   |
        M=  |   0   M   |   ,K= |   E-C  D-B  |
        such that scipy.linalg.eig(K,M) solves the problem 

        THE SYSTEM MUST BE LINEARISED FOR THIS TO WORK
        """
        M_prime = sym.eye(self.qs*2)
        M_prime[-self.qs:,-self.qs:]=self.M

        _Q = self.ExtForces.Q() if self.ExtForces is not None else sym.Matrix([0]*self.qs)

        K_prime = sym.zeros(self.qs*2)
        K_prime[:self.qs,-self.qs:] = sym.eye(self.qs)
        K_prime[-self.qs:,:self.qs] = _Q.jacobian(self.q)-self.f.jacobian(self.q)
        K_prime[-self.qs:,-self.qs:] = _Q.jacobian(self.qd)-self.f.jacobian(self.qd)

        return K_prime, M_prime

    @staticmethod
    def _jacobian(M,x):
        return sym.Matrix([[*M.diff(xi)] for xi in x]).T

    def to_file(self,p,filename):
        #Get string represtations
        M_code = "def get_M(p):\n\t"+print_model(self.M,p).replace('\n','\n\t')+"\n\treturn e\n"
        f_code = "def get_f(p):\n\t"+print_model(self.f,p).replace('\n','\n\t')+"\n\treturn e\n"
        T_code = "def get_T(p):\n\t"+print_model(self.T,p).replace('\n','\n\t')+"\n\treturn e\n"
        U_code = "def get_U(p):\n\t"+print_model(self.U,p).replace('\n','\n\t')+"\n\treturn e\n"
        p_code = 'def get_p():\n\t'+p.print_python().replace('\n','\n\t')+"\n\treturn p\n"

        if self.ExtForces is not None:
            Q_code = "def get_Q(p):\n\t"+print_model(self.ExtForces.Q(),p).replace('\n','\n\t')+"\n\treturn e\n"
        else:
            Q_code = "def get_Q(p):\n\t"+"return ImmutableDenseMatrix([[0]"+",[0]"*(self.M.shape[0]-1)+"])\n"
        if self.C is not None:
            C_code = 'def get_C(p):\n\t'+print_model(self.C,p).replace('\n','\n\t')+"\n\treturn e\n"
        else:
            C_code = 'def get_C(p):\n\treturn None\n'
        #Combine and add import statements
        full_code = "from sympy import *\nimport moyra as ma\n\n"+M_code+f_code+T_code+U_code+Q_code+C_code+p_code

        # Save to the file
        t_file = open(filename,"w")
        n = t_file.write(full_code)
        t_file.close()   


    @classmethod
    def from_file(cls,filename):
        import importlib.util
        from .forces import ExternalForce
        spec = importlib.util.spec_from_file_location("my.Model", filename)
        m = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(m)
        p = m.get_p()
        M = m.get_M(p)
        f = m.get_f(p)
        T = m.get_T(p)
        U = m.get_U(p)
        _Q = m.get_Q(p)
        C = m.get_C(p)
        ExtForce = ExternalForce(_Q)
        return (cls(M,f,T,U,ExtForce,C),p)


    def to_matlab_class(self,p,file_dir,class_name,base_class = None,additional_funcs = [], octave_user_functions={}):
        funcs = [('get_M',self.M),('get_f',self.f),('get_Q',self.ExtForces.Q()),
                ('get_KE',self.T),('get_PE',self.U)]
        funcs = [*funcs,*additional_funcs]
        if self.C is not None:
            funcs.append(('get_C',self.C))
            C_q = sym.simplify(self.C).jacobian(self.q)
            C_t =  partial_wrt_t(self.C,p.q)
            C_tt = partial_wrt_t(C_t,p.q)
            C_qt = partial_wrt_t(C_q,p.q)
            Q_c = C_tt + (C_q*p.qd).jacobian(p.q)*p.qd + 2*C_qt*p.qd
            M_lag = sym.BlockMatrix([[self.M,C_q.T],[C_q,sym.zeros(len(self.C))]]).as_explicit()
            Q_lag = sym.BlockMatrix([[self.f-self.ExtForces.Q()],[Q_c]]).as_explicit()

            funcs.append(('get_C_q',C_q))
            funcs.append(('get_C_t',C_t))
            # funcs.append(('get_C_tt',me.msubs(C_tt))
            funcs.append(('get_Q_c',Q_c))
            funcs.append(('get_M_lag',M_lag))
            funcs.append(('get_Q_lag',Q_lag))
        # create directory
        class_dir = os.path.join(file_dir,f"@{class_name}")
        if not os.path.exists(class_dir):
            os.mkdir(class_dir)
        for func_name,matrix in funcs:
            with open(os.path.join(class_dir,f"{func_name}.m"),'w') as file:
                file.write(self._gen_octave(matrix,func_name,octave_user_functions))
        p.to_matlab_class(class_name=class_name, file_dir=class_dir, base_class=base_class )
    
    def to_matlab_file(self,p,file_dir,octave_user_functions={}):
        funcs = (('get_M',self.M),('get_f',self.f),('get_Q',self.ExtForces.Q()))
        for func_name,matrix in funcs:
            with open(os.path.join(file_dir,f"{func_name}.m"),'w') as file:
                file.write(self._gen_octave(matrix,func_name,octave_user_functions))
        p.to_matlab_class(file_dir=file_dir)

    def to_matlab_file_linear(self,p,file_dir,octave_user_functions={}):
        mats = self.extract_matrices()
        names = ['A','B','C','D','E']
        funcs = list(zip([f'get_{i}' for i in names],mats))
        for func_name,matrix in funcs:
            with open(os.path.join(file_dir+f"{func_name}.m"),'w') as file:
                file.write(self._gen_octave(matrix,func_name,octave_user_functions))
        p.to_matlab_class(file_dir=file_dir)
        
    def _gen_octave(self,expr,func_name, user_functions={}):
        # convert states to non-time dependent variable
        U = sym.Matrix(sym.symbols(f'u_:{self.qs*2}'))
        l = dict(zip([*self.q,*self.qd],U))
        l_deriv = dict(zip(self.q.diff(t),self.qd))
        expr = me.msubs(expr,l_deriv)
        expr = me.msubs(expr,l)

        # get parameter replacements
        param_string = '%% extract required parameters from structure\n\t'
        matries = []
        unknown_vars = []
        for var in expr.free_symbols:
            if isinstance(var,ModelValue):
                if isinstance(var,ModelMatrixSymbol):
                    if var._matrix not in matries:
                        param_string += f'{var._matrix} = p.{var._matrix};\n\t'
                        matries.append(var._matrix)
                elif isinstance(var,ModelSymbol):
                    param_string += f'{var.name} = p.{var.name};\n\t'
                elif isinstance(var,ModelMatrix):
                    param_string += f'{var._matrix_symbol} = p.{var._matrix_symbol};\n\t'
            elif var not in U:
                print(f'Unknown variable {var} found in function {func_name}. It will be added to the function signature.')
                unknown_vars.append(var)


        # split expr into groups
        replacments, exprs = sym.cse(expr,symbols=(sym.Symbol(f'rep_{i}')for i in range(10000)))
        if isinstance(expr,tuple):
            expr = tuple(exprs)
        elif isinstance(expr,list):
            expr = exprs
        else:
            expr = exprs[0]      

        group_string = '%% create common groups\n\t'
        for variable, expression in replacments:
            group_string +=f'{variable} = {sym.printing.octave.octave_code(expression, user_functions=user_functions)};\n\t'
        
        # convert to octave string and covert states to vector form
        out = '%% create output vector\n\tout = ' + sym.printing.octave.octave_code(expr, user_functions=user_functions)

        #convert state vector calls
        my_replace = lambda x: f'U({int(x.group(1))+1})'
        out = re.sub(r"u_(?P<index>\d+)",my_replace,out)
        group_string = re.sub(r"u_(?P<index>\d+)",my_replace,group_string)

        # make the file pretty...
        out = out.replace(',',',...\n\t\t').replace(';',';...\n\t\t')

        file_sig = f'%{func_name.upper()} Auto-generated function from moyra\n\t'
        file_sig += f'%\n\t'
        file_sig += f'%\tCreated at : {ctime(time())} \n\t'
        file_sig += f'%\tCreated with : moyra https://pypi.org/project/moyra/\n\t'
        file_sig += f'%\n\t'


        # wrap output in octave function signature
        # create unknow var string
        unknown_str = '' if not unknown_vars else ','+','.join(sorted([str(i) for i in unknown_vars], key=str))
        signature = f'function out = {func_name}(p,U{unknown_str})\n\t'
        

        ## tidy params with curly braces in name
        my_replace = lambda x: f'_{x.group(1)}'
        param_string = re.sub(r"_\{(?P<index>.+)\}",my_replace,param_string)
        group_string = re.sub(r"_\{(?P<index>.+)\}",my_replace,group_string)
        out = re.sub(r"_\{(?P<index>.+)\}",my_replace,out)

        octave_string = signature + file_sig + param_string + group_string + out + ';\nend'
        return octave_string






